#include "CyanBus.hpp"

// C++ system level
// #include <cstdlib>  // TODO: is this for memcpy?
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "esp_log.h"
#include "driver/gpio.h"
#include "driver/uart.h"
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"

// project specific
#include "LambdaTask.hpp"

// qthing stuff
// #include ""



static const char* TAG = "CyanBus";


namespace SiliconTorch {

  namespace CyanBus {

    // This leads to ~54 P/s @ 2 MBaud at 100% load
    const uint32_t MTU = 4096;

    const uint8_t* const HEADER = (const uint8_t*)"fxCyan";


    CyanBus::CyanBus(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate, uint8_t uartChannel) : tx(tx), rx(rx), de(de), re(re), uartChannel(uartChannel) {


      uint64_t bitMask = 0;

      bitMask |= 1 << tx;
      bitMask |= 1 << de;
      bitMask |= 1 << re;

      gpio_config_t gpio = {
        .pin_bit_mask = bitMask,
        .mode         = GPIO_MODE_OUTPUT,
        .pull_up_en   = GPIO_PULLUP_DISABLE,
        .pull_down_en = GPIO_PULLDOWN_DISABLE,
        .intr_type    = GPIO_INTR_DISABLE,        
      };

      // configure outputs
      gpio_config(&gpio);
      txrxEN(false);

      gpio.pin_bit_mask = 1 << rx;
      gpio.mode = GPIO_MODE_INPUT;

      // configure inputs
      gpio_config(&gpio);


      uart_config_t uartConfig = {
        .baud_rate           = (int)baudRate,
        .data_bits           = UART_DATA_8_BITS,
        .parity              = UART_PARITY_DISABLE,
        .stop_bits           = UART_STOP_BITS_1,
        .flow_ctrl           = UART_HW_FLOWCTRL_DISABLE,
        .rx_flow_ctrl_thresh = 0,
        .use_ref_tick        = false,
        // .source_clk = UART_SCLK_APB,
      };


      uart_port_t _ch = (uart_port_t)uartChannel;

      uart_driver_install(_ch, MTU * 2, 0, 32, &uartEvQ, 0);  // txBuf = 0; queue size = 32;  TODO: evaluate!!!
      uart_param_config(_ch, &uartConfig);
      uart_set_pin(_ch, tx, rx, UART_PIN_NO_CHANGE, UART_PIN_NO_CHANGE);
      uart_set_mode(_ch, UART_MODE_UART);
      uart_set_rx_timeout(_ch, 0);  // TODO: timeout…??

      // TOOD: receive only ATM!
      txEN(false);
      rxEN(true);


      readerTaskHandle = new SiliconTorch::Util::LambdaTask([&](){ readerTask(); });
    }

    void CyanBus::readerTask() {

      uart_event_t event;

      uint32_t bufLEN = MTU * 1.2;
      uint8_t* buffer = new uint8_t[bufLEN];
      uint8_t* bufPTR = buffer;

      ESP_LOGW(TAG, "bufLEN = %d   BufferSize[MTU * 1.2] = %d", bufLEN, (int)sizeof(buffer));



    }


    void CyanBus::setBaudRate(uint32_t baudRate) {
      uart_set_baudrate((uart_port_t)uartChannel, baudRate);
    }

    uint32_t CyanBus::getBaudRate() {
      uint32_t baudRate = 0;
      uart_get_baudrate((uart_port_t)uartChannel, &baudRate);
      return baudRate;
    }


    void CyanBus::txEN(bool state) {
      gpio_set_level((gpio_num_t)de, state);
    }

    void CyanBus::rxEN(bool state) {
      gpio_set_level((gpio_num_t)re, state ^ 1);
    }
    
    void CyanBus::txrxEN(bool state) {
      txEN(state);
      rxEN(state);
    }

  }
}



