#pragma once

// C++ system level
#include <string>
#include <cinttypes>
#include <functional>

// ESP32 specific
#include "freertos/FreeRTOS.h"
#include "freertos/queue.h"

// project specific
#include "LambdaTask.hpp"


namespace SiliconTorch {

  namespace CyanBus {

    // CyanBus payload MTU
    extern const uint32_t MTU;

    // CyanBus envelope header
    extern const uint8_t* const HEADER;

    class CyanBus {
      public:
        CyanBus(uint8_t tx, uint8_t rx, uint8_t de, uint8_t re, uint32_t baudRate = 115200, uint8_t uartChannel = 1);

        uint32_t getBaudRate();
        void setBaudRate(uint32_t baudRate);

      private:

        QueueHandle_t uartEvQ;

        uint8_t uartChannel;
        uint8_t tx, rx, de, re;

        void txEN(bool state);
        void rxEN(bool state);
        void txrxEN(bool state);


        // Receive only ATM!  (make public after sound implementation…)
        void write(std::string& data);
        void write(const uint8_t* data, uint32_t length);




        void readerTask();

        SiliconTorch::Util::LambdaTask* readerTaskHandle = NULL;

    };
  }
}
